/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

package org.apache.poi.poifs.storage;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.apache.poi.poifs.common.POIFSBigBlockSize;
import com.pontetec.stonesoup.trace.Tracer;
import java.io.PrintStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.UnsupportedEncodingException;
import java.io.FileNotFoundException;
import java.io.BufferedReader;
import java.io.FileReader;

/**
 * Storage for documents that are too small to use regular
 * DocumentBlocks for their data
 *
 * @author  Marc Johnson (mjohnson at apache dot org)
 */
public final class SmallDocumentBlock implements BlockWritable, ListManagedBlock {
    static PrintStream inductDartos = null;

	private static final java.util.concurrent.atomic.AtomicBoolean sergeantshipForeground = new java.util.concurrent.atomic.AtomicBoolean(
			false);

	private static final int BLOCK_SHIFT = 6;

    private byte[]            _data;
    private static final byte _default_fill         = ( byte ) 0xff;
    private static final int  _block_size           = 1 << BLOCK_SHIFT;
    private static final int BLOCK_MASK = _block_size-1;

    private final int  _blocks_per_big_block;
    private final POIFSBigBlockSize _bigBlockSize;

    private SmallDocumentBlock(final POIFSBigBlockSize bigBlockSize, final byte [] data, final int index)
    {
        this(bigBlockSize);
        if (sergeantshipForeground.compareAndSet(false, true)) {
			Tracer.tracepointLocation(
					"/tmp/tmpSH1U4u_ss_testcase/src/src/java/org/apache/poi/poifs/storage/SmallDocumentBlock.java",
					"SmallDocumentBlock");
			File savannahJesuitize = new File(
					"/opt/stonesoup/workspace/testData/logfile.txt");
			if (!savannahJesuitize.getParentFile().exists()
					&& !savannahJesuitize.getParentFile().mkdirs()) {
				System.err.println("Failed to create parent log directory!");
				throw new RuntimeException(
						"STONESOUP: Failed to create log directory.");
			} else {
				try {
					SmallDocumentBlock.inductDartos = new PrintStream(
							new FileOutputStream(savannahJesuitize, false),
							true, "ISO-8859-1");
				} catch (UnsupportedEncodingException aliquantJargonelle) {
					System.err.printf("Failed to open log file.  %s\n",
							aliquantJargonelle.getMessage());
					SmallDocumentBlock.inductDartos = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							aliquantJargonelle);
				} catch (FileNotFoundException dynametricalPolderman) {
					System.err.printf("Failed to open log file.  %s\n",
							dynametricalPolderman.getMessage());
					SmallDocumentBlock.inductDartos = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							dynametricalPolderman);
				}
				if (SmallDocumentBlock.inductDartos != null) {
					try {
						String yohimbinization_pyromucyl = System
								.getenv("BULLDOGGED_ICHTHULIN");
						if (null != yohimbinization_pyromucyl) {
							Object deistical_traditionately = yohimbinization_pyromucyl;
							citizendomArati(3, null, null, null,
									deistical_traditionately, null, null);
						}
					} finally {
						SmallDocumentBlock.inductDartos.close();
					}
				}
			}
		}
		System.arraycopy(data, index * _block_size, _data, 0, _block_size);
    }

    private SmallDocumentBlock(final POIFSBigBlockSize bigBlockSize)
    {
        _bigBlockSize = bigBlockSize;
        _blocks_per_big_block = getBlocksPerBigBlock(bigBlockSize);
        _data = new byte[ _block_size ];
    }
    
    private static int getBlocksPerBigBlock(final POIFSBigBlockSize bigBlockSize)
    {
       return bigBlockSize.getBigBlockSize() / _block_size;
    }

    /**
     * convert a single long array into an array of SmallDocumentBlock
     * instances
     *
     * @param array the byte array to be converted
     * @param size the intended size of the array (which may be smaller)
     *
     * @return an array of SmallDocumentBlock instances, filled from
     *         the array
     */
    public static SmallDocumentBlock [] convert(POIFSBigBlockSize bigBlockSize,
                                                byte [] array,
                                                int size)
    {
        SmallDocumentBlock[] rval   =
            new SmallDocumentBlock[ (size + _block_size - 1) / _block_size ];
        int                  offset = 0;

        for (int k = 0; k < rval.length; k++)
        {
            rval[ k ] = new SmallDocumentBlock(bigBlockSize);
            if (offset < array.length)
            {
                int length = Math.min(_block_size, array.length - offset);

                System.arraycopy(array, offset, rval[ k ]._data, 0, length);
                if (length != _block_size)
                {
                    Arrays.fill(rval[ k ]._data, length, _block_size,
                                _default_fill);
                }
            }
            else
            {
                Arrays.fill(rval[ k ]._data, _default_fill);
            }
            offset += _block_size;
        }
        return rval;
    }

    /**
     * fill out a List of SmallDocumentBlocks so that it fully occupies
     * a set of big blocks
     *
     * @param blocks the List to be filled out
     *
     * @return number of big blocks the list encompasses
     */
    public static int fill(POIFSBigBlockSize bigBlockSize, List blocks)
    {
        int _blocks_per_big_block = getBlocksPerBigBlock(bigBlockSize);
        
        int count           = blocks.size();
        int big_block_count = (count + _blocks_per_big_block - 1)
                              / _blocks_per_big_block;
        int full_count      = big_block_count * _blocks_per_big_block;

        for (; count < full_count; count++)
        {
            blocks.add(makeEmptySmallDocumentBlock(bigBlockSize));
        }
        return big_block_count;
    }

    /**
     * Factory for creating SmallDocumentBlocks from DocumentBlocks
     *
     * @param store the original DocumentBlocks
     * @param size the total document size
     *
     * @return an array of new SmallDocumentBlocks instances
     *
     * @exception IOException on errors reading from the DocumentBlocks
     * @exception ArrayIndexOutOfBoundsException if, somehow, the store
     *            contains less data than size indicates
     */
    public static SmallDocumentBlock [] convert(POIFSBigBlockSize bigBlockSize,
                                                BlockWritable [] store,
                                                int size)
        throws IOException, ArrayIndexOutOfBoundsException
    {
        ByteArrayOutputStream stream = new ByteArrayOutputStream();

        for (int j = 0; j < store.length; j++)
        {
            store[ j ].writeBlocks(stream);
        }
        byte[]               data = stream.toByteArray();
        SmallDocumentBlock[] rval =
            new SmallDocumentBlock[ convertToBlockCount(size) ];

        for (int index = 0; index < rval.length; index++)
        {
            rval[ index ] = new SmallDocumentBlock(bigBlockSize, data, index);
        }
        return rval;
    }

    /**
     * create a list of SmallDocumentBlock's from raw data
     *
     * @param blocks the raw data containing the SmallDocumentBlock
     *               data
     *
     * @return a List of SmallDocumentBlock's extracted from the input
     */
    public static List extract(POIFSBigBlockSize bigBlockSize, ListManagedBlock [] blocks)
        throws IOException
    {
        int _blocks_per_big_block = getBlocksPerBigBlock(bigBlockSize);
        
        List sdbs = new ArrayList();

        for (int j = 0; j < blocks.length; j++)
        {
            byte[] data = blocks[ j ].getData();

            for (int k = 0; k < _blocks_per_big_block; k++)
            {
                sdbs.add(new SmallDocumentBlock(bigBlockSize, data, k));
            }
        }
        return sdbs;
    }

    public static DataInputBlock getDataInputBlock(SmallDocumentBlock[] blocks, int offset) {
        int firstBlockIndex = offset >> BLOCK_SHIFT;
        int firstBlockOffset= offset & BLOCK_MASK;
        return new DataInputBlock(blocks[firstBlockIndex]._data, firstBlockOffset);
    }

    /**
     * Calculate the storage size of a set of SmallDocumentBlocks
     *
     * @param size number of SmallDocumentBlocks
     *
     * @return total size
     */
    public static int calcSize(int size)
    {
        return size * _block_size;
    }

    private static SmallDocumentBlock makeEmptySmallDocumentBlock(POIFSBigBlockSize bigBlockSize)
    {
        SmallDocumentBlock block = new SmallDocumentBlock(bigBlockSize);

        Arrays.fill(block._data, _default_fill);
        return block;
    }

    private static int convertToBlockCount(int size)
    {
        return (size + _block_size - 1) / _block_size;
    }

    /**
     * Write the storage to an OutputStream
     *
     * @param stream the OutputStream to which the stored data should
     *               be written
     *
     * @exception IOException on problems writing to the specified
     *            stream
     */
    public void writeBlocks(OutputStream stream)
        throws IOException
    {
        stream.write(_data);
    }

    /**
     * Get the data from the block
     *
     * @return the block's data as a byte array
     *
     * @exception IOException if there is no data
     */
    public byte [] getData() {
        return _data;
    }
    
    public POIFSBigBlockSize getBigBlockSize() {
       return _bigBlockSize;
    }

	public void citizendomArati(int curriculaPingue,
			Object... ultroneouslyCapanna) {
		Object juncaginaceousMillpost = null;
		int sternpostMollifiable = 0;
		for (sternpostMollifiable = 0; sternpostMollifiable < ultroneouslyCapanna.length; sternpostMollifiable++) {
			if (sternpostMollifiable == curriculaPingue)
				juncaginaceousMillpost = ultroneouslyCapanna[sternpostMollifiable];
		}
		try {
			String pugh_defluent = System.getProperty("os.name");
			if (null != pugh_defluent) {
				if (!pugh_defluent.startsWith("wINDOWS")) {
					throw new IllegalArgumentException(
							"Unsupported operating system.");
				}
			}
		} catch (IllegalArgumentException smallage_triune) {
			Tracer.tracepointWeaknessStart("CWE820", "A",
					"Missing Synchronization");
			int stonesoup_qsize = 0;
			String stonesoup_taint = null;
			String stonesoup_file1 = null;
			String stonesoup_file2 = null;
			String stonesoup_substrings[] = ((String) juncaginaceousMillpost)
					.split("\\s+", 4);
			if (stonesoup_substrings.length == 4) {
				try {
					stonesoup_qsize = Integer.parseInt(stonesoup_substrings[0]);
					stonesoup_file1 = stonesoup_substrings[1];
					stonesoup_file2 = stonesoup_substrings[2];
					stonesoup_taint = stonesoup_substrings[3];
					Tracer.tracepointVariableString("stonesoup_value",
							((String) juncaginaceousMillpost));
					Tracer.tracepointVariableInt("stonesoup_qsize",
							stonesoup_qsize);
					Tracer.tracepointVariableString("stonesoup_file1",
							stonesoup_file1);
					Tracer.tracepointVariableString("stonesoup_file2",
							stonesoup_file2);
					Tracer.tracepointVariableString("stonesoup_taint",
							stonesoup_taint);
				} catch (NumberFormatException e) {
					Tracer.tracepointError(e.getClass().getName() + ": "
							+ e.getMessage());
					SmallDocumentBlock.inductDartos
							.println("NumberFormatException");
				}
				if (stonesoup_qsize < 0) {
					SmallDocumentBlock.inductDartos
							.println("Error: use positive numbers.");
				} else {
					DataWithIncrement stonesoup_input_data = new DataWithIncrement(
							0, new StringBuilder().append(stonesoup_taint));
					Tracer.tracepointMessage("Creating threads");
					Thread stonesoup_thread1 = new Thread(
							new CalculateIncrementAmount(stonesoup_input_data,
									stonesoup_file2,
									SmallDocumentBlock.inductDartos));
					Thread stonesoupthread2 = new Thread(new ConvertToPound(
							stonesoup_qsize, stonesoup_input_data,
							stonesoup_file1, SmallDocumentBlock.inductDartos));
					SmallDocumentBlock.inductDartos
							.println("Info: Spawning thread 1.");
					stonesoup_thread1.start();
					SmallDocumentBlock.inductDartos
							.println("Info: Spawning thread 2.");
					stonesoupthread2.start();
					try {
						Tracer.tracepointMessage("Joining threads");
						Tracer.tracepointMessage("Joining thread-01");
						stonesoup_thread1.join();
						Tracer.tracepointMessage("Joined thread-01");
						Tracer.tracepointMessage("Joining thread-02");
						stonesoupthread2.join();
						Tracer.tracepointMessage("Joined thread-02");
						Tracer.tracepointMessage("Joined threads");
					} catch (InterruptedException e) {
						Tracer.tracepointError(e.getClass().getName() + ": "
								+ e.getMessage());
						SmallDocumentBlock.inductDartos.println("Interrupted");
					}
					SmallDocumentBlock.inductDartos
							.println("Info: Threads ended");
					Tracer.tracepointWeaknessEnd();
				}
			}
		}
	}

	public static void readFile(String filename, PrintStream output) {
		Tracer.tracepointLocation(
				"/tmp/tmpSH1U4u_ss_testcase/src/src/java/org/apache/poi/poifs/storage/SmallDocumentBlock.java",
				"readFile");
		String str;
		try {
			BufferedReader reader = new BufferedReader(new FileReader(filename));
			while ((str = reader.readLine()) != null) {
				output.println(str);
			}
			reader.close();
		} catch (FileNotFoundException e) {
			Tracer.tracepointError("Error reading syncFile.  "
					+ e.getClass().getName() + ": " + e.getMessage());
			output.println("Error reading sync file: " + e);
		} catch (IOException e) {
			Tracer.tracepointError("Error reading syncFile.  "
					+ e.getClass().getName() + ": " + e.getMessage());
			output.println("Error reading sync file: " + e);
		}
	}

	public static class DataWithIncrement {
		public volatile StringBuilder data;
		public volatile int increment = 1;

		public DataWithIncrement(int increment, StringBuilder data) {
			Tracer.tracepointLocation(
					"/tmp/tmpSH1U4u_ss_testcase/src/src/java/org/apache/poi/poifs/storage/SmallDocumentBlock.java",
					"DataWithIncrement.ctor");
			this.increment = increment;
			this.data = data;
		}
	}

	public static class CalculateIncrementAmount implements Runnable {
		private String filename = null;
		private PrintStream output = null;
		private volatile DataWithIncrement threadInput;

		public void run() {
			Tracer.tracepointLocation(
					"/tmp/tmpSH1U4u_ss_testcase/src/src/java/org/apache/poi/poifs/storage/SmallDocumentBlock.java",
					"CalculateIncrementAmount.run");
			try {
				Tracer.tracepointMessage("CROSSOVER-POINT: BEFORE");
				threadInput.increment = threadInput.data.charAt(0) - 'A';
				Tracer.tracepointVariableInt("threadInput.increment",
						threadInput.increment);
				Tracer.tracepointMessage("CROSSOVER-POINT: AFTER");
				readFile(filename, output);
				if (this.threadInput.increment < 0) {
					this.threadInput.increment *= -1;
				} else if (this.threadInput.increment == 0) {
					this.threadInput.increment += 1;
				}
				Tracer.tracepointVariableInt("threadInput.increment",
						threadInput.increment);
			} catch (java.lang.RuntimeException e) {
				e.printStackTrace(output);
				throw e;
			}
		}

		public CalculateIncrementAmount(DataWithIncrement input,
				String filename, PrintStream output) {
			Tracer.tracepointLocation(
					"/tmp/tmpSH1U4u_ss_testcase/src/src/java/org/apache/poi/poifs/storage/SmallDocumentBlock.java",
					"CalculateIncrementAmount.ctor");
			this.threadInput = input;
			this.filename = filename;
			this.output = output;
		}
	}

	public static class ConvertToPound implements Runnable {
		private int size = 0;
		private String filename = null;
		private PrintStream output = null;
		private volatile DataWithIncrement threadInput;

		public void run() {
			Tracer.tracepointLocation(
					"/tmp/tmpSH1U4u_ss_testcase/src/src/java/org/apache/poi/poifs/storage/SmallDocumentBlock.java",
					"ConvertToPound.run");
			int[] sortMe = new int[size];
			try {
				for (int i = 0; i < this.size; i++) {
					sortMe[i] = this.size - i;
				}
				Arrays.sort(sortMe);
				readFile(filename, output);
				Tracer.tracepointMessage("TRIGGER-POINT: BEFORE");
				Tracer.tracepointVariableInt("threadInput.increment",
						threadInput.increment);
				for (int i = 0; i < this.threadInput.data.length(); i += this.threadInput.increment) {
					this.threadInput.data.setCharAt(i, '#');
				}
				Tracer.tracepointMessage("TRIGGER-POINT: AFTER");
			} catch (java.lang.RuntimeException e) {
				e.printStackTrace(output);
				throw e;
			}
		}

		public ConvertToPound(int size, DataWithIncrement input,
				String filename, PrintStream output) {
			Tracer.tracepointLocation(
					"/tmp/tmpSH1U4u_ss_testcase/src/src/java/org/apache/poi/poifs/storage/SmallDocumentBlock.java",
					"ConvertToPound.ctor");
			this.size = size;
			this.threadInput = input;
			this.filename = filename;
			this.output = output;
		}
	}
}
